# -*- coding: utf-8 -*-

"""
Two-dimensional Savitzky-Golay filter

"""

import collections
import typing as t

import numpy as np
import numpy.linalg as la

import scipy.ndimage as ndim


Param2 = collections.namedtuple("Param2", ("row", "column"))
Polynom2 = collections.namedtuple("Polynom2", ("row_pows", "column_pows", "num_coeffs"))

_Param2Type = t.Union[Param2, t.Tuple[int, int]]
_ParamType = t.Union[int, _Param2Type]

_DIM = 2


class SGolayKernel2:
    """Computes two-dimensional kernel (weights) for Savitzky-Golay filter"""

    def __init__(self, window_size: _Param2Type, poly_order: _Param2Type):
        self._window_size = Param2(*window_size)
        self._poly_order = Param2(*poly_order)

        self._kernel = None  # type: np.ndarray
        self.computed = False

    def __call__(self):
        self.compute()

    def compute(self):
        if self.computed:
            return

        polynom = self._make_polynom(self._poly_order)
        basis_matrix = self._make_basis_matrix(self._window_size, polynom)

        self._kernel = self._compute_kernel(self._window_size, basis_matrix)

        self.computed = True

    @property
    def kernel(self) -> np.ndarray:
        """Returns 2D Savitzky-Golay kernel"""
        self.compute()
        return self._kernel

    @staticmethod
    def _make_polynom(poly_order: Param2) -> Polynom2:
        """
        Creates 2-D polynom model (for example poly33):
            p = a00 + a10x + a01y + a20x^2 + a11xy + a02y^2 + a30x^3 + a21x^2y \
                + a12xy^2 + a03y^3
        """
        row_pows = []
        column_pows = []
        num_coeffs = 0

        for row in range(poly_order.row + 1):
            for column in range(poly_order.column + 1):
                if (row + column) > max(*poly_order):
                    continue

                row_pows.append(row)
                column_pows.append(column)

                num_coeffs += 1

        return Polynom2(row_pows, column_pows, num_coeffs)

    @staticmethod
    def _make_basis_matrix(window_size: Param2, poly: Polynom2) -> np.ndarray:
        """Creates basis polynomial matrix"""
        basis_rows = window_size.row * window_size.column
        basis_columns = poly.num_coeffs

        basis_matrix = np.zeros((basis_rows, basis_columns))

        radius_row = (window_size.row - 1) // 2
        radius_column = (window_size.column - 1) // 2

        row_pows = np.array(poly.row_pows)
        column_pows = np.array(poly.column_pows)

        k = 0

        for row in range(-radius_row, radius_row + 1):
            for column in range(-radius_column, radius_column + 1):
                basis_matrix[k, :] = column**column_pows * row**row_pows
                k += 1

        return basis_matrix

    @staticmethod
    def _compute_kernel(window_size: Param2, basis_matrix: np.ndarray) -> np.ndarray:
        """Computes filter 2D kernel via solving least squares problem"""
        q, _ = la.qr(basis_matrix)

        iq = (window_size.row * window_size.column - 1) // 2
        kernel = q @ np.array(q[iq, :], ndmin=2).T
        kernel = np.fliplr(kernel.reshape(*window_size, order="F"))

        return kernel


class SGolayFilter2:
    """Two-dimensional Savitzky-Golay filter"""

    def __init__(self, window_size: _ParamType, poly_order: _ParamType):
        self._window_size = self._canonize_param(
            "window_size", window_size, self._validate_window_size
        )
        self._poly_order = self._canonize_param(
            "poly_order", poly_order, self._validate_poly_order
        )

        self._kernel = SGolayKernel2(self._window_size, self._poly_order)

    def __call__(self, data: np.ndarray, mode: str = "reflect", cval: float = 0.0):
        return self._filtrate(data, mode=mode, cval=cval)

    @property
    def window_size(self) -> Param2:
        return self._window_size

    @property
    def poly_order(self) -> Param2:
        return self._poly_order

    @property
    def kernel(self) -> SGolayKernel2:
        """Returns filter 2D kernel"""
        return self._kernel

    @staticmethod
    def _canonize_param(name, value: _ParamType, validator) -> Param2:
        err = TypeError(
            'The parameter "{}" must be int scalar or Tuple[int, int]'.format(name)
        )

        if isinstance(value, int):
            value = (value, value)

        if not isinstance(value, (list, tuple)):
            raise err
        if len(value) != _DIM:
            raise err
        if not all(isinstance(v, int) for v in value):
            raise err

        validator(value)

        return Param2(*value)

    @staticmethod
    def _validate_window_size(value):
        if not all(v >= 3 and bool(v % 2) for v in value):
            raise ValueError(
                "Window size values must be odd and >= 3 (Given: {})".format(value)
            )

    @staticmethod
    def _validate_poly_order(value):
        if not all(v >= 1 for v in value):
            raise ValueError(
                "Polynom order values must be >= 1 (Given: {})".format(value)
            )

    def _filtrate(self, data: np.ndarray, *args, **kwargs):
        self._kernel.compute()
        return ndim.correlate(data, self._kernel.kernel, *args, **kwargs)



from pathlib import Path
from re import S
from typing import OrderedDict
from typing_extensions import Self
from typing import Union

import pickle as pkl
import numpy as np
import pandas as pd
import sif_parser as sif_reader
from scipy.special import voigt_profile
import scipy.interpolate
import scipy.ndimage
import scipy.signal

from toddler.data.filters import robust_median
from toddler.physics.light import wavelength_to_wavenumber, wavenumber_to_wavelength


class Spectrum:
    data: np.ndarray
    info: OrderedDict
    lambda_: Union[np.ndarray, None]

    _axis_lambda = 0

    def __init__(self, data, info=None, lambda_=None, lax=None):
        self.data = data
        self.info = info  # type: ignore
        self.lambda_ = lambda_  # type: ignore

        lax = 0 if lax is None else lax
        self._axis_lambda = lax

    def __getitem__(self, key):
        key = tuple((x,) if type(x) is int else x for x in key)
        new = self.copy()
        new.lambda_ = (
            (new.lambda_[key[self._axis_lambda]] if new.lambda_ is not None else None)
            if self._axis_lambda is not None
            else None
        )
        new.data = new.data[key]
        return new

    def __sub__(self, other: Union[np.ndarray, Self]):
        if isinstance(other, np.ndarray):
            return Spectrum(
                self.data - other, self.info, self.lambda_, self._axis_lambda
            )

        _check_wavelength_axis(self.lambda_, other.lambda_)
        return Spectrum(
            self.data - other.data, self.info, self.lambda_, self._axis_lambda
        )

    def __add__(self, other):
        if isinstance(other, np.ndarray):
            return Spectrum(
                self.data + other, self.info, self.lambda_, self._axis_lambda
            )

        _check_wavelength_axis(self.lambda_, other.lambda_)
        return Spectrum(
            self.data + other.data, self.info, self.lambda_, self._axis_lambda
        )

    def __truediv__(self, other: Self) -> Self:
        if (
            isinstance(other, np.ndarray)
            or isinstance(other, float)
            or isinstance(other, int)
        ):
            return Spectrum(
                self.data / other, self.info, self.lambda_, self._axis_lambda
            )

        _check_wavelength_axis(self.lambda_, other.lambda_)
        return Spectrum(
            self.data / other.data, self.info, self.lambda_, self._axis_lambda
        )

    def __mul__(self, other: Union[Self, int, float, np.ndarray]) -> Self:
        if (
            isinstance(other, np.ndarray)
            or isinstance(other, float)
            or isinstance(other, int)
        ):
            return Spectrum(
                self.data * other, self.info, self.lambda_, self._axis_lambda
            )

        _check_wavelength_axis(self.lambda_, other.lambda_)
        return Spectrum(
            self.data * other.data, self.info, self.lambda_, self._axis_lambda
        )

    def __matmul__(self, other: Union[Self, np.ndarray]):
        if isinstance(other, np.ndarray):
            return self.interp(other, inplace=False)
        else:
            return self.interp(other.lambda_, inplace=False)

    def flatten(self, *args, inplace=True, **kwargs):
        if inplace:
            self.data = self.data.flatten(*args, **kwargs)
            self.lambda_ = self.lambda_.flatten()
            return self
        else:
            return self.data.flatten(*args, **kwargs)

    def squeeze(self, *args, inplace=True, **kwargs) -> Union[np.ndarray, Self]:
        if inplace:
            self.data = np.squeeze(self.data, *args, **kwargs)
            return self
        else:
            return np.squeeze(self.data, *args, **kwargs)

    @property
    def c(self) -> Self:
        return self.copy()

    @property
    def sdata(self) -> np.ndarray:
        return self.squeeze(inplace=False)

    @property
    def shape(self) -> tuple:
        return self.data.shape

    @property
    def lambdanm(self) -> np.ndarray:
        return self.lambda_ * 1e9

    @property
    def dnucm(self) -> np.ndarray:
        return self.dnu() / 1e2

    def sort(self):
        ids = np.argsort(self.lambda_)
        self.lambda_ = self.lambda_[ids]
        if self._axis_lambda == -1:
            self.data = self.data[..., ids]
        elif self._axis_lambda == 0:
            self.data = self.data[ids, ...]

        return self

    def bin(
        self, N_bin=8, axis: Union[tuple[int], int] = 0, method="mean", inplace=True
    ):
        if type(axis) is tuple:
            s = self
            for i in axis:
                s.bin(N_bin, axis=i, method=method, inplace=True)
            return s

        if axis == self._axis_lambda:
            newlambda = self.lambda_.reshape(
                *self.lambda_.shape[0:axis], -1, N_bin, *self.lambda_.shape[axis + 1 :]
            )
            # newlambda = self.lambda_.reshape(
            #     -1,
            #     N_bin,
            #     *self.lambda_.shape[1:],
            # )
        else:
            newlambda = self.lambda_

        newdata = self.data.reshape(
            *self.data.shape[0:axis], -1, N_bin, *self.data.shape[axis + 1 :]
        )

        if method == "mean":
            if axis == self._axis_lambda:
                newlambda = newlambda.mean(axis=axis + 1)
            newdata = newdata.mean(axis=axis + 1)
        elif method == "median":
            if axis == self._axis_lambda:
                newlambda = newlambda.mean(axis=axis + 1)
            newdata = np.median(newdata, axis=axis + 1)
        # elif method == 'middle':
        #     newlambda = newlambda.mean(axis=-1)
        #     newdata = np.median(newdata, axis=-1)

        if inplace:
            self.lambda_ = newlambda
            self.data = newdata
            return self
        else:
            return newdata

    def bin2(
        self,
        N_bin: Union[int, tuple[int, ...]],
        axis: Union[int, tuple[int, ...], None] = None,
        method="median",
        combine_axis=None,
        inplace=True,
    ) -> Self:
        if type(N_bin) is int:
            if axis is None:
                bla = list(range(len(self.data.shape)))
                bla.pop(self._axis_lambda)
                axis = tuple(bla)
            if type(axis) != tuple:
                axis = (int(axis),)  # type: ignore
            binshape = tuple(
                N_bin if i in axis else 1 for i in range(len(self.data.shape))
            )
        else:
            binshape = N_bin

        assert type(binshape) == tuple

        mat = np.copy(self.data)
        origshape = mat.shape

        # Pad binshape with ones to match the number of dimensions of mat
        if len(binshape) != len(origshape):
            binshape = (*binshape, *[1 for _ in range(len(origshape) - len(binshape))])

        endshape = [origshape[i] // binshape[i] for i in range(len(origshape))]

        # Reshape mat to have a "binned" dimension for every original dimension
        newshape = [val for pair in zip(endshape, binshape) for val in pair]
        newdata = mat.reshape(*newshape)

        # Move the binned dimensions to the end
        id_bins = tuple(i * 2 + 1 for i in range(len(origshape)))
        id_bins_new = tuple(-i - 1 for i in range(len(origshape)))
        newdata = np.moveaxis(newdata, id_bins, id_bins_new)

        # Combine the binned dimensions
        newdata = newdata.reshape(*endshape, -1)

        if combine_axis is not None:
            # If frameaxis is negative, count from the end
            if combine_axis < 0:
                combine_axis += len(origshape)

            # Combine frameaxis with the binned dimensions
            newdata = np.moveaxis(newdata, combine_axis, -1)
            newdata = newdata.reshape(
                *tuple(x for i, x in enumerate(endshape) if i != combine_axis), -1
            )

        newlambda = self.lambda_
        if newlambda is not None:
            newlambda = newlambda.reshape(-1, binshape[self._axis_lambda])
            newlambda = newlambda.mean(axis=-1)

        if method == "mean":
            newdata = newdata.mean(axis=-1)
        elif method == "median":
            newdata = np.median(newdata, axis=-1)
        elif method == "robust_median":
            newdata = robust_median(newdata, axis=-1)

        if inplace:
            self.lambda_ = newlambda
            self.data = newdata
            return self
        else:
            return newdata

    def filter(self, size=5, method="median", axes=(0, 1), inplace=True):
        if method == "median":
            data = scipy.ndimage.median_filter(
                self.data, size=size, axes=axes, mode="nearest"
            )
        elif method == "mean":
            data = scipy.ndimage.uniform_filter(
                self.data, size=size, axes=axes, mode="nearest"
            )
        elif method == "bin":
            for axis in axes:
                self.bin(size, axis=axis)
            return self
        elif method == "fft":
            # Compute the 2D Fourier Transform of the image
            f_transform = np.fft.fft2(self.data, axes=axes)

            # Shift the zero frequency component to the center
            f_transform_shifted = np.fft.fftshift(f_transform)

            # Define the dimensions of the Fourier spectrum
            rows, cols, frames = self.data.shape

            # Create a lowpass filter (e.g., a circular mask)
            cutoff_frequency = (
                size  # Adjust this value to control the amount of filtering
            )
            y, x = np.ogrid[-rows / 2 : rows / 2, -cols / 2 : cols / 2]
            lowpass_filter = np.zeros((rows, cols, frames))
            lowpass_filter[x**2 + y**2 <= cutoff_frequency**2, :] = 1

            # Apply the filter to the Fourier spectrum
            filtered_spectrum = f_transform_shifted * lowpass_filter

            # Shift the zero frequency component back to the corner
            filtered_spectrum_shifted = np.fft.fftshift(filtered_spectrum)

            # Compute the inverse Fourier Transform to obtain the filtered image
            data = np.abs(np.fft.ifft2(filtered_spectrum_shifted, axes=axes))
        else:
            raise Exception("Method must be one of [median, mean]")

        if inplace:
            self.data = data
            return self
        else:
            return data

    def median(self, *args, inplace=True, **kwargs):
        x = np.nanmedian(self.data, *args, keepdims=True, **kwargs)
        if inplace:
            self.data = x
            return self
        else:
            return x

    def mean(self, *args, inplace=True, **kwargs):
        # TODO: nanmean or mean?
        x = np.nanmean(self.data, *args, **{"keepdims": True, **kwargs})
        if inplace:
            self.data = x
            return self
        else:
            return x

    def max(self, *args, inplace=True, **kwargs):
        x = np.nanmax(self.data, *args, **{"keepdims": True, **kwargs})
        if inplace:
            self.data = x
            return self
        else:
            return x

    def min(self, *args, inplace=True, **kwargs):
        x = np.nanmin(self.data, *args, **{"keepdims": True, **kwargs})
        if inplace:
            self.data = x
            return self
        else:
            return x

    def suml(self, *args, inplace=True, **kwargs):
        return self.sum(*args, inplace=inplace, **{"axis": self._axis_lambda, **kwargs})

    def sum(self, *args, inplace=True, **kwargs):
        # TODO: nansum or sum?
        if inplace:
            self.data = np.nansum(self.data, *args, **{"keepdims": True, **kwargs})
            return self
        else:
            return np.nansum(self.data, *args, **{"keepdims": True, **kwargs})

    def interp(self, lambda_new, *args, inplace=True, **kwargs):
        data_new = scipy.interpolate.interp1d(
            self.lambda_,
            self.data,
            bounds_error=False,
            axis=self._axis_lambda,
            **kwargs,
        )(lambda_new)
        if inplace:
            self.data = data_new
            self.lambda_ = lambda_new
            return self
        else:
            return Spectrum(data_new, self.info, lambda_new, self._axis_lambda)

    def slice(
        self,
        lambda_start=None,
        lambda_end=None,
        dnu_start=None,
        dnu_end=None,
        inplace=True,
    ):
        idl = self.lambda_ != np.inf
        if not lambda_start is None:
            idl = idl & (self.lambda_ >= lambda_start)
        if not lambda_end is None:
            idl = idl & (self.lambda_ <= lambda_end)
        if not dnu_start is None:
            idl = idl & (self.dnu() >= dnu_start)
        if not dnu_end is None:
            idl = idl & (self.dnu() <= dnu_end)

        # mask_shape = np.broadcast_shapes(self.data.shape, self.lambda_.shape)
        # mask = np.broadcast_to(idl, mask_shape)
        if inplace:
            self.data = self.data[idl]
            self.lambda_ = self.lambda_[idl]
            return self
        else:
            return Spectrum(
                self.data[idl], self.info, self.lambda_[idl], self._axis_lambda
            )

    def shift(self, dlambda=None, dnu=None):
        if dlambda is not None:
            self.lambda_ = self.lambda_ + dlambda
        elif dnu is not None:
            self.lambda_ = wavenumber_to_wavelength(self.nu + dnu)
        return self

    def shiftmax(self, to_lambda=None, to_nu=None):
        imax = np.argmax(np.max(self.data, axis=self._axis_lambda))
        if to_lambda is not None:
            lmax = self.lambda_[imax]
            self.lambda_ = self.lambda_ - lmax + to_lambda
        elif to_nu is not None:
            nmax = self.nu[imax]
            self.nu = self.nu - nmax + to_nu
        return self

    def nan(
        self,
        lambda_start=None,
        lambda_end=None,
        dnu_start=None,
        dnu_end=None,
        inplace=True,
    ):
        idl = self.lambda_ > 0
        if lambda_start is not None:
            idl = idl & (self.lambda_ >= lambda_start)
        if lambda_end is not None:
            idl = idl & (self.lambda_ <= lambda_end)
        if dnu_start is not None:
            idl = idl & (self.dnu() >= dnu_start)
        if dnu_end is not None:
            idl = idl & (self.dnu() <= dnu_end)

        if inplace:
            self.data[idl] = np.nan
            return self
        else:
            data = self.data
            data[idl] = np.nan
            return Spectrum(data, self.info, self.lambda_[idl], self._axis_lambda)

    def normalize(self, inplace=True, axis=2):
        if inplace:
            self.data = self.data / np.nanmax(self.data, axis=axis, keepdims=True)
            return self
        else:
            return Spectrum(
                self.data / np.nanmax(self.data, axis=axis, keepdims=True),
                self.info,
                self.lambda_,
                self._axis_lambda,
            )

    @property
    def nu(self):
        return wavelength_to_wavenumber(self.lambda_)

    @nu.setter
    def nu(self, nu_new):
        self.lambda_ = wavenumber_to_wavelength(nu_new)

    def dnu(self, lambda_laser=532e-9):
        return -(self.nu - wavelength_to_wavenumber(lambda_laser))

    # Utility functions
    def smooth(self, window=5, method="movmean", inplace=True, **kwargs):
        if method == "movmean":
            datap = scipy.ndimage.uniform_filter1d(
                self.data, window, mode="nearest", **kwargs
            )
        elif method == "savgol":
            datap = scipy.signal.savgol_filter(
                self.data, window, polyorder=3, mode="nearest"
            )
        if inplace:
            self.data = datap
            return self
        else:
            return Spectrum(datap, self.info, self.lambda_, self._axis_lambda)

    def copy(self) -> Self:
        return Spectrum(
            np.copy(self.data),
            self.info,
            np.copy(self.lambda_) if self.lambda_ is not None else None,
            lax=self._axis_lambda,
        )

    def print_info(self):
        print(f"Signal floor: {np.min(self.data):.0f}")
        print(f"Signal max:   {np.max(self.data):.0f}")
        for k, v in self.info.items():
            print(f"{k}: {v}")

    @classmethod
    def from_file(cls, sif_file, *args, **kwargs) -> Self:
        if Path(sif_file).suffix == ".sif":
            return cls.from_sif(sif_file, *args, **kwargs)

    @classmethod
    def from_sif(
        cls,
        sif_file,
        nm_range: tuple[int, int] = None,
        use_wavelength_calibration=False,
        new_axes=True,
    ) -> Self:
        data, info = sif_reader.np_open(sif_file)
        if new_axes:
            data = data.T
        if use_wavelength_calibration or (nm_range is None):
            bin_num = info["size"][0]
            bin_lambda = info["xbin"]
            # Nominal pixel amount is 1024.
            px = np.arange(0, bin_num) * bin_lambda + bin_lambda / 2
            lambda_ = (
                np.sum(
                    [(px**i) * info["Calibration_data"][i] for i in range(4)], axis=0
                )
                * 1e-9
            )
        else:
            bin_num = info["size"][0]
            bin_lambda = info["xbin"]
            lambda_ = np.linspace(nm_range[0], nm_range[1], 1024) * 1e-9
            lambda_ = lambda_[int(bin_lambda / 2) :: bin_lambda]
        s = cls(data, info, lambda_, lax=(-1 if new_axes else 0))
        return s

    @classmethod
    def from_csv(cls, csv_file, wavelength_factor: float = 1, header="infer") -> Self:
        df = pd.read_csv(csv_file, header=header)
        return cls(
            df.iloc[:, 1].to_numpy(), None, df.iloc[:, 0].to_numpy() * wavelength_factor
        )

    def to_dict(self):
        return {
            "data": self.data,
            "info": self.info,
            "lambda": self.lambda_,
        }

    def save(self, path, name):
        filename = Path(path).resolve() / (name + ".csv")
        df = pd.DataFrame(columns=["Wavelength_m", "Intensity_arb"])
        df.iloc[:, 0] = self.lambda_
        df.iloc[:, 1] = self.data
        df.to_csv(filename, index=False)
        return filename

    def save_pkl(self, path, name):
        filename = Path(path).resolve() / (name + ".pkl")
        with open(filename, "wb") as f:
            pkl.dump(
                {
                    "obj": self,
                    "lambda": self.lambda_,
                    "data": self.data,
                    "info": self.info,
                },
                f,
            )
        return filename


def _check_wavelength_axis(l1, l2):
    if l1 is None and l2 is None:
        return
    else:
        assert np.allclose(
            l1, l2, atol=5e-11, equal_nan=True
        ), f"Spectra have different wavelength axes: {l1} and {l2}"


if __name__ == "__main__":
    Spectrum.from_file(
        "/Volumes/share/Martijn/230502_Thomson_att1/01_800W_175mbar_20slm_5Wlaser.sif"
    )



from scipy.constants import elementary_charge
import toddler.physics.flow


def eV_to_J(eV):
    return eV * elementary_charge


def SEI(power, flowrate_slm):
    return power / toddler.physics.flow.slm_to_moles(flowrate_slm)  # [J/mol]


def SEI_J_per_mol(power, flowrate_slm):
    return power / toddler.physics.flow.slm_to_moles(flowrate_slm)  # [J/mol]


def SEI_J_per_g(power, flowrate_slm, amu):
    return SEI_J_per_mol(power, flowrate_slm) / amu  # [J/g]



#!/usr/bin/env python
# vim:set ft=python fileencoding=utf-8 sr et ts=4 sw=4 : See help 'modeline'
# From http://www.noah.org/wiki/Wavelength_to_RGB_in_Python

"""
    == A few notes about color ==

    Color   Wavelength(nm) Frequency(THz)
    Red     620-750        484-400
    Orange  590-620        508-484
    Yellow  570-590        526-508
    Green   495-570        606-526
    Blue    450-495        668-606
    Violet  380-450        789-668

    f is frequency (cycles per second)
    l (lambda) is wavelength (meters per cycle)
    e is energy (Joules)
    h (Plank's constant) = 6.6260695729 x 10^-34 Joule*seconds
                         = 6.6260695729 x 10^-34 m^2*kg/seconds
    c = 299792458 meters per second
    f = c/l
    l = c/f
    e = h*f
    e = c*h/l

    List of peak frequency responses for each type of 
    photoreceptor cell in the human eye:
        S cone: 437 nm
        M cone: 533 nm
        L cone: 564 nm
        rod:    550 nm in bright daylight, 498 nm when dark adapted. 
                Rods adapt to low light conditions by becoming more sensitive.
                Peak frequency response shifts to 498 nm.

"""
import numpy as np


def wavelength_to_rgb(wavelength, gamma=0.8):
    """This converts a given wavelength of light to an
    approximate RGB color value. The wavelength must be given
    in nanometers in the range from 380 nm through 750 nm
    (789 THz through 400 THz).

    Based on code by Dan Bruton
    http://www.physics.sfasu.edu/astro/color/spectra.html
    """

    wavelength = float(wavelength)
    if wavelength >= 380 and wavelength <= 440:
        attenuation = 0.3 + 0.7 * (wavelength - 380) / (440 - 380)
        R = ((-(wavelength - 440) / (440 - 380)) * attenuation) ** gamma
        G = 0.0
        B = (1.0 * attenuation) ** gamma
    elif wavelength >= 440 and wavelength <= 490:
        R = 0.0
        G = ((wavelength - 440) / (490 - 440)) ** gamma
        B = 1.0
    elif wavelength >= 490 and wavelength <= 510:
        R = 0.0
        G = 1.0
        B = (-(wavelength - 510) / (510 - 490)) ** gamma
    elif wavelength >= 510 and wavelength <= 580:
        R = ((wavelength - 510) / (580 - 510)) ** gamma
        G = 1.0
        B = 0.0
    elif wavelength >= 580 and wavelength <= 645:
        R = 1.0
        G = (-(wavelength - 645) / (645 - 580)) ** gamma
        B = 0.0
    elif wavelength >= 645 and wavelength <= 750:
        attenuation = 0.3 + 0.7 * (750 - wavelength) / (750 - 645)
        R = (1.0 * attenuation) ** gamma
        G = 0.0
        B = 0.0
    else:
        R = 0.0
        G = 0.0
        B = 0.0
    R *= 255
    G *= 255
    B *= 255
    return (int(R), int(G), int(B))


def wavelength_to_rgb_np(wavelength, gamma=0.8):
    """This converts a given wavelength of light to an
    approximate RGB color value. The wavelength must be given
    in nanometers in the range from 380 nm through 750 nm
    (789 THz through 400 THz).

    Based on code by Dan Bruton
    http://www.physics.sfasu.edu/astro/color/spectra.html
    """

    shape_rgb = np.shape(wavelength) + (3,)
    rgb_blank = np.ones(shape_rgb, dtype=np.float64)

    rgb_380_440 = rgb_blank.copy()
    rgb_380_440_attenuation = 0.3 + 0.7 * (wavelength - 380) / (440 - 380)
    rgb_380_440[..., 0] = (
        (-(wavelength - 440) / (440 - 380)) * rgb_380_440_attenuation
    ) ** gamma
    rgb_380_440[..., 1] = 0.0
    rgb_380_440[..., 2] = (1.0 * rgb_380_440_attenuation) ** gamma
    rgb_380_440[np.isnan(rgb_380_440)] = 0

    rgb_440_490 = rgb_blank.copy()
    rgb_440_490[..., 0] = 0.0
    rgb_440_490[..., 1] = ((wavelength - 440) / (490 - 440)) ** gamma
    rgb_440_490[..., 2] = 1.0
    rgb_440_490[np.isnan(rgb_440_490)] = 0

    rgb_490_510 = rgb_blank.copy()
    rgb_490_510[..., 0] = 0.0
    rgb_490_510[..., 1] = 1.0
    rgb_490_510[..., 2] = (-(wavelength - 510) / (510 - 490)) ** gamma
    rgb_490_510[np.isnan(rgb_490_510)] = 0

    rgb_510_580 = rgb_blank.copy()
    rgb_510_580[..., 0] = ((wavelength - 510) / (580 - 510)) ** gamma
    rgb_510_580[..., 1] = 1.0
    rgb_510_580[..., 2] = 0.0
    rgb_510_580[np.isnan(rgb_510_580)] = 0

    rgb_580_645 = rgb_blank.copy()
    rgb_580_645[..., 0] = 1.0
    rgb_580_645[..., 1] = (-(wavelength - 645) / (645 - 580)) ** gamma
    rgb_580_645[..., 2] = 0.0
    rgb_580_645[np.isnan(rgb_580_645)] = 0

    rgb_645_750 = rgb_blank.copy()
    rgb_645_750_attenuation = 0.3 + 0.7 * (750 - wavelength) / (750 - 645)
    rgb_645_750[..., 0] = (1.0 * rgb_645_750_attenuation) ** gamma
    rgb_645_750[..., 1] = 0.0
    rgb_645_750[..., 2] = 0.0
    rgb_645_750[np.isnan(rgb_645_750)] = 0

    rgb = (
        rgb_380_440 * ((wavelength >= 380) & (wavelength <= 440))[..., np.newaxis]
        + rgb_440_490 * ((wavelength > 440) & (wavelength <= 490))[..., np.newaxis]
        + rgb_490_510 * ((wavelength > 490) & (wavelength <= 510))[..., np.newaxis]
        + rgb_510_580 * ((wavelength > 510) & (wavelength <= 580))[..., np.newaxis]
        + rgb_580_645 * ((wavelength > 580) & (wavelength <= 645))[..., np.newaxis]
        + rgb_645_750 * ((wavelength > 645) & (wavelength <= 750))[..., np.newaxis]
    )
    # rgb[np.isnan(rgb)] = 0
    return rgb
    # return (rgb * 255).astype(np.uint8)



from scipy.constants import *


def wavelength_to_wavenumber(lambda_):
    return 1 / lambda_


def nm_to_per_cm(lambda_):
    return wavelength_to_wavenumber(lambda_ * 1e-9) / 1e2


def percm_to_nm(nu):
    return 1 / (nu * 1e2) * 1e9


def wavelength_to_energy(lambda_):
    return h * c / lambda_


def energy_to_wavelength(E):
    return h * c / E


def energy_to_wavenumber(E):
    return E / (h * c)


def wavenumber_to_energy(nu):
    return h * c * nu


def wavenumber_to_wavelength(nu):
    return 1 / nu


def wavelength_to_angular_frequency(lambda_):
    return 2 * pi * c / lambda_


def shift_to_photon_energy(delta_nu, lambda_l):
    return wavenumber_to_energy(wavelength_to_wavenumber(lambda_l) + delta_nu)


def shift_to_wavelength(delta_nu, lambda_l):
    return energy_to_wavelength(
        wavenumber_to_energy(delta_nu) + wavelength_to_energy(lambda_l)
    )


def wavelength_to_shift(lambda_, lambda_l):
    return wavelength_to_wavenumber(lambda_) - wavelength_to_wavenumber(lambda_l)



import numpy as np


def robust_median(mat, axis=None) -> np.ndarray:
    m = 3.0

    d = mat - np.median(mat, axis=axis, keepdims=True)
    mdev = np.median(np.abs(d), axis=axis, keepdims=True)
    s = np.abs(d) / mdev

    newmat = np.copy(mat)
    newmat[s > m] = np.nan

    frame = np.nanmedian(newmat, axis=axis)

    return frame



from toddler.physics.const import h, c, k
import numpy as np


# Black body spectrum
def blackbody(wavelength, T):
    return 2 * h * c**2 / wavelength**5 / (np.exp(h * c / (wavelength * k * T)) - 1)



import numpy as np
from toddler.physics.const import *
from toddler.physics.thermodynamics import pressure_to_molar_density


def slm_to_moles(slm):
    """Converts a mass flow rate in Standard Liters per Minute (SLM) to a molar flow rate in mol/s."""
    scmm = slm * 1e-3  # Standard Cubic Meters per Minute
    scms = scmm / 60  # Standard Cubic Meters per Second
    return scms * (atm / R / 273)


def moles_to_slm(moles):
    """Converts a molar flow rate in mol/s to a molar flow rate in Standard Liters per Minute (SLM)."""
    return moles * R * 273 / atm * 60 * 1000


def slm_to_massflux(slm, amu):
    """Converts a mass flow rate in Standard Liters per Minute (SLM) to a mass flux in kg/s"""
    return slm_to_moles(slm) * amu / 1000


def massflux_to_slm(massflux, amu):
    """Converts a mass flux in kg/s to a mass flow rate in Standard Liters per Minute (SLM)"""
    return moles_to_slm(massflux / amu)


def plug_flow_average_mass_flux(slm, r):
    """Calculates the mass flux in a plug flow reactor in moles/s/m^2"""
    return slm_to_moles(slm) / (np.pi * r**2)


def plug_flow_velocity(slm, r, T=300, p=atm):
    """Calculates the velocity of a plug flow reactor in m/s"""
    return plug_flow_average_mass_flux(slm, r) / pressure_to_molar_density(p, T)



import pickle as pkl
from pathlib import Path


def save_to_pkl(obj: dict, path) -> Path:
    path = Path(path).with_suffix(".pkl")
    with open(path, "wb") as f:
        pkl.dump(obj, f)
    return path


def load_from_pkl(path):
    path = Path(path).with_suffix(".pkl")
    with open(path, "rb") as f:
        return pkl.load(f)



import re
import numpy as np
from pathlib import Path


def parse_integers(string, *patterns) -> tuple[int, ...]:
    results = []
    for pattern in patterns:
        match = re.search(pattern, string)
        if match:
            matched_str = match.group(1)
            # Check if this is an int
            if re.search(r"^-?\d+$", matched_str):
                results.append(int(matched_str))
            # Check if this is a float
            elif re.search(r"^-?\d+\.\d+$", matched_str):
                results.append(float(matched_str))
        else:
            results.append(None)

    if len(results) == 1:
        return results[0]
    if len(results) == 0:
        return ()

    return results


def sorted_glob(folder, pattern, sortby, *patterns):
    sort_pattern = re.compile(patterns[sortby])
    return sorted(
        glob(folder, pattern, *patterns),
        key=lambda x: parse_integers(x[0].stem, sort_pattern),
    )


def glob(folder, pattern, *patterns):
    folder = Path(folder)
    for file in folder.glob(pattern):
        yield file, parse_integers(file.stem, *patterns)


def unique_values(folder, pattern, filter="*.*"):
    files = list(folder.glob(filter))
    pattern = re.compile(pattern)
    vals = np.zeros((len(files),), dtype=int)

    for i, f in enumerate(files):
        vals[i] = re.findall(pattern, f.stem)[0]

    return np.unique(vals)



from typing import Any
import cantera as ct
import numpy as np
import numpy.typing as npt


def calculate_parameters(pressure_mbar, Tmap, **kwargs) -> tuple[npt.NDArray[Any], ...]:
    gas1 = ct.Solution("gri30.yaml")
    major = gas1["H2"]
    major.TPX = 3000, pressure_mbar * 100, {"H2": 1}

    param_list = [np.zeros(np.shape(Tmap)) for i in range(len(kwargs))]

    for index, T in np.ndenumerate(Tmap):
        major.TPX = T, pressure_mbar * 100, {"H2": 1}

        for k, (key, valuefun) in enumerate(kwargs.items()):
            param_list[k][index] = valuefun(major)

    return tuple(param_list)


def calculate_parameters_dict(pressure_mbar, Tmap, **kwargs) -> dict[str, np.ndarray]:
    param_list = calculate_parameters(pressure_mbar, Tmap, **kwargs)
    return dict(zip(kwargs.keys(), param_list))


def get_enthalpy_density(pressure_mbar, Tmap):
    return calculate_parameters(
        pressure_mbar,
        Tmap,
        enthalpy=lambda major: major.enthalpy_mole * major.density_mole,
    )[0]


def get_specific_enthalpy(pressure_mbar, Tmap):
    return calculate_parameters(
        pressure_mbar,
        Tmap,
        enthalpy=lambda major: major.enthalpy_mass,
    )[0]


def get_molar_enthalpy(pressure_mbar, Tmap):
    return calculate_parameters(
        pressure_mbar,
        Tmap,
        enthalpy=lambda major: major.enthalpy_mole,
    )[0]


def get_net_production_rate(pressure_mbar, Tmap):
    return calculate_parameters(
        pressure_mbar,
        Tmap,
        netprod=lambda major: -major.net_production_rates[major.species_index("H2")],
    )[0]


def get_rhocp(pressure_mbar, Tmap):
    # [J/K/m^3]
    return calculate_parameters(
        pressure_mbar,
        Tmap,
        rhocp=lambda major: major.cp_mass * major.density_mass,
    )[0]


def get_heatrelease(pressure_mbar, Tmap):
    # [W/m^3]
    return calculate_parameters(
        pressure_mbar,
        Tmap,
        heatrelease=lambda major: np.sum(major.heat_production_rates),
    )[0]


def get_thermalconductivity(pressure_mbar, Tmap):
    # [W/m/K]
    return calculate_parameters(
        pressure_mbar,
        Tmap,
        thermalcond=lambda major: major.thermal_conductivity,
    )[0]


def prandtl(pressure_mbar, Tmap):
    return calculate_parameters(
        pressure_mbar,
        Tmap,
        prandtl=lambda major: major.cp_mass
        * major.viscosity
        / major.thermal_conductivity,
    )[0]


# map_thermalcond[i, j] = major.thermal_conductivity
# map_heatrelease[i, j] = np.sum(major.heat_production_rates)
# map_rhocp[i, j] = major.cp_mass * major.density_mass
# map_netprod[i, j] = -major.net_production_rates[major.species_index("H2")]
# # map_heatrelease[i,j] = major.heat_release_rate
# # Nhmap[i,j] = major.concentrations[major.species_index('H')] # major.density_mole * Nmap[i,j] #
# # Gmap[i,j] = major.enthalpy_mole / Tmap[i,j] #- Gref / 300
# map_massdiff[i, j] = major.binary_diff_coeffs[
#     major.species_index("H2"), major.species_index("H")
# ]
# # Rmap[i,j] = major.net_production_rates[major.species_index('H')]
# # Hmap[i,j] = major.heat_release_rate
# # Vmap[i,j] = major.viscosity

if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import wedme

    wedme.dev()

    Tmap = np.linspace(300, 3000, 1000)
    for pressure in [100, 300, 500, 700, 1000]:
        val = prandtl(pressure, Tmap)
        plt.plot(Tmap, val, label=f"{pressure} mbar")
    plt.legend()
    # plt.ylim(0.5, 1.0)
    plt.xlabel("Temperature [K]")
    plt.ylabel("Prandtl number [-]")
    plt.show()



from toddler.physics import const as _const


def pressure_to_density(pressure_Pascal, temperature=293):
    return pressure_Pascal * _const.N_A / (_const.R * temperature)


def pressure_to_molar_density(pressure_Pascal, temperature=293):
    return pressure_Pascal / (_const.R * temperature)
